/*
 * Driver for LML SFPGA SPI Controllers
 *
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/clk.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/interrupt.h>
#include <linux/spi/spi.h>

#include "lmlsfpga_spi.h"
#include "support_fpga_hw.h"

//#define DEBUG_GENERAL_LMLSFPGA 1      /* Enable to know general program flow */
//#define DEBUG_LMLSFPGA_SPI_DRV 1      /* Enable for sysctl interface/debug RX bufs */
//#define DBG_CAN_INTE_INTF_EFLG 1      /* Enable for understanding CANINTE/CANINTF handling */

#ifdef DEBUG_LMLSFPGA_SPI_DRV 
#define MAX_DBG_ARRAY_SIZE 512
#define MAX_RX_BUF_SIZE 14

//struct dbg_can_rx_data_t { 
unsigned long rx_pckt_cnt;
unsigned int rx_array_index;
//unsigned int rx_frame[MAX_DBG_ARRAY_SIZE][MAX_RX_BUF_SIZE];
//unsigned int rx_bus_num[MAX_DBG_ARRAY_SIZE];

typedef struct {
	u8 *rx_frame;
	unsigned int rx_bus_num;
}frame_t;

frame_t *rx_pckt_defn = NULL;
//};
#endif

struct lmlsfpga_spi {
	spinlock_t		lock;
	struct platform_device	*pdev;
	struct spi_transfer *transfer;  /* current transfer */
};


#ifdef DEBUG_LMLSFPGA_SPI_DRV 

void lmlsfpga_spi_clear_debug_info(void) 
{
	unsigned int pckt_indx = 0, byte_indx = 0;

	if (rx_pckt_defn) {
		for ( pckt_indx = 0 ; pckt_indx < MAX_DBG_ARRAY_SIZE; pckt_indx++ ) {
			//rx_bus_num[pckt_indx] = 0;
			rx_pckt_defn[pckt_indx].rx_bus_num = 0;
			for ( byte_indx = 0 ; byte_indx < MAX_RX_BUF_SIZE; byte_indx++ ) {
				if (rx_pckt_defn[pckt_indx].rx_frame) {
					rx_pckt_defn[pckt_indx].rx_frame[byte_indx] = 0;
				}
			//rx_frame[pckt_indx][byte_indx] = 0;
			}
		}
	}
	rx_array_index = 0;
}


void lmlsfpga_spi_print_debug_info(void) 
{
    unsigned int pckt_indx = 0;

	printk(KERN_ALERT "RX packet count = %lu\n", rx_pckt_cnt);
	printk(KERN_ALERT "Current array index = %u\n\n", rx_array_index); 
    printk(KERN_ALERT " +------+-----+-------------+\n") ;
    printk(KERN_ALERT " | Indx | bus | rx_data_buf |\n") ;
    printk(KERN_ALERT " +------+-----+-------------+\n") ;
    
	if (rx_pckt_defn) {
		for ( pckt_indx = 0 ; pckt_indx < MAX_DBG_ARRAY_SIZE; pckt_indx++ ) {
			printk(KERN_ALERT "%6u | %5u | %x %x %x %x %x %x %x %x %x %x %x %x %x %x\n",
					pckt_indx,
					rx_pckt_defn[pckt_indx].rx_bus_num, 
					rx_pckt_defn[pckt_indx].rx_frame[0],
					rx_pckt_defn[pckt_indx].rx_frame[1],
					rx_pckt_defn[pckt_indx].rx_frame[2],
					rx_pckt_defn[pckt_indx].rx_frame[3],
					rx_pckt_defn[pckt_indx].rx_frame[4],
					rx_pckt_defn[pckt_indx].rx_frame[5],
					rx_pckt_defn[pckt_indx].rx_frame[6],
					rx_pckt_defn[pckt_indx].rx_frame[7],
					rx_pckt_defn[pckt_indx].rx_frame[8],
					rx_pckt_defn[pckt_indx].rx_frame[9],
					rx_pckt_defn[pckt_indx].rx_frame[10],
					rx_pckt_defn[pckt_indx].rx_frame[11],
					rx_pckt_defn[pckt_indx].rx_frame[12],
					rx_pckt_defn[pckt_indx].rx_frame[13]);
		}
    }
    printk(KERN_ALERT " +-----+-------------+\n") ;
}

static ssize_t
show_dump(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
    ssize_t ret = 0 ;
    lmlsfpga_spi_print_debug_info() ;
    ret = sprintf(buf,"%s\n"," ==== LMLSFPGA_SPI Debug Info ==== ") ;
    return ret ;
}

static ssize_t
show_clear_debug_data(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
    ssize_t ret = 0 ;
    lmlsfpga_spi_clear_debug_info() ;
    ret = sprintf(buf,"%s\n"," ==== LMLSFPGA_SPI Cleared Info ==== ") ;
    return ret ;
}

static struct kobj_attribute lmlsfpga_spi_attributes[] = {
    __ATTR(dump,                 S_IRUGO,            show_dump,             NULL ),
    __ATTR(clear_debug_data,     S_IRUGO,            show_clear_debug_data,       NULL ),
} ;

static struct kobject *lmlsfpga_spi_dir_obj;

int lmlsfpga_spi_sysfs_create_node_and_attrs(char *node_name, struct kobject **node, struct kobject *parent, struct kobj_attribute attr_array[], unsigned int attr_arr_size )
{
    int retval = 0 ;
    int i = 0 ;

    if ( ( node_name == NULL ) || ( node == NULL ) || ( parent == NULL ) )
        return -EINVAL ;

    *node = kobject_create_and_add(node_name, parent);
    if (*node) {
        for ( i=0; i < attr_arr_size/sizeof(struct kobj_attribute) ; i++ ) {
            retval = sysfs_create_file(*node, &attr_array[i].attr);
            if ( !retval) { // retval = 0 is success.
                // printk("(S)::%s() %s attr %s add.\n", __FUNCTION__, node_name, attr_array[i].attr.name ) ;
            } else {
                // Error.
                printk("(F)::%s() %s attr %s add.\n", __FUNCTION__, node_name, attr_array[i].attr.name ) ;
                return retval ; // break ; // Break from for loop.
            }
        }
    } else {
        printk("(F)::%s() %s kobject_create_and_add \n", __FUNCTION__, node_name ) ;
        retval = -ENOMEM ;
    }
    return retval ;
}

int lmlsfpga_spi_sysfs_init(void)
{
    int retval = 0, pckt_indx = 0;

	printk(KERN_ALERT " LMLSFPGA_SPI Sys Fs Init \n ");

	rx_pckt_defn = kzalloc((sizeof(frame_t)) * MAX_DBG_ARRAY_SIZE, GFP_ATOMIC);
	if (!rx_pckt_defn) {
		printk (KERN_ALERT "%s(): kzalloc failed to allocate memory to frame_t array of structures of length %u",__FUNCTION__, MAX_DBG_ARRAY_SIZE);
		return -1;
	}
	for (pckt_indx = 0; pckt_indx<MAX_DBG_ARRAY_SIZE; pckt_indx++) {
		rx_pckt_defn[pckt_indx].rx_frame = kzalloc((sizeof(u8)) * MAX_RX_BUF_SIZE, GFP_ATOMIC);
		if (!(rx_pckt_defn[pckt_indx].rx_frame)) {
			printk (KERN_ALERT "%s(): kzalloc failed to allocate memory to array of u8 of length %u",__FUNCTION__, MAX_RX_BUF_SIZE);
			kfree(rx_pckt_defn);
			return -1;
		}
	}

    // First register switch node under /sys/kernel.
    retval = lmlsfpga_spi_sysfs_create_node_and_attrs("lmlsfpga_spi_debug", &lmlsfpga_spi_dir_obj, kernel_kobj, &lmlsfpga_spi_attributes[0], sizeof(lmlsfpga_spi_attributes) ) ;
    if ( !retval )  {
        printk (KERN_ALERT " %s() sw_sysfs_create_node_and_attrs passed \n", __FUNCTION__ ) ;
    } else {
        // Error.
        printk(KERN_ALERT " %s() switch kobject_create_and_add failure \n", __FUNCTION__ ) ;
        return retval ;
    }
    return 0;
}

void lmlsfpga_spi_sysfs_exit(void)
{
	int retval = 0, pckt_indx = 0;
	
	printk(KERN_ALERT " LMLSFPGA_SPI Sys Fs Exit \n ");
	
	for (pckt_indx = 0; pckt_indx < MAX_DBG_ARRAY_SIZE; pckt_indx++) {
		if (rx_pckt_defn[pckt_indx].rx_frame) {
			kfree(rx_pckt_defn[pckt_indx].rx_frame);
		}
	}
	if (rx_pckt_defn) {
		kfree(rx_pckt_defn);
	}
}
#endif

int supfpga_mm_reg16_write(uint16_t spi_msg, unsigned long reg_addr ){
	return support_fpga_write_parallel_bus_reg((uint32_t)spi_msg, (uint32_t)reg_addr);
}

uint16_t supfpga_mm_reg16_read(unsigned long reg_addr) {
    uint32_t reg_value;
    int ret;
    ret = support_fpga_read_parallel_bus_reg(&reg_value, (uint32_t)reg_addr);
    return (uint16_t)(reg_value);
}


static int lmlsfpga_spi_setup(struct spi_device *spi)
{
	int			ret;
	uint16_t reg_value;
	unsigned long spi_ie_n, spi_len_offset;

	switch(spi->master->bus_num) {
		case 3://SPI Controller 0
			spi_len_offset = SPI_0_LEN; //SPI 0 controller and length offset
			spi_ie_n = SPI_IE_0; //SPI 0 interrupt controller enable bit
			break;
		case 4://SPI Controller 1
			spi_len_offset = SPI_1_LEN;
			spi_ie_n = SPI_IE_1;
			break;
		case 5://SPI Controller 2
			spi_len_offset = SPI_2_LEN;
			spi_ie_n = SPI_IE_2;
			break;
		case 6://SPI Controller 3
			spi_len_offset = SPI_3_LEN;
			spi_ie_n = SPI_IE_3;
			break;
		default:
			printk("Unknown controller\n");
			return -1;
	}

	reg_value = (((uint16_t)SPI_SIE) << 8);
	/*Write to the SPI_N_CTRL register to initialize the SPI controller*/
	ret = supfpga_mm_reg16_write(reg_value, spi_len_offset); //16 bit write
	if (ret != 0) {
		printk("lmlsfpga_spi_setup(): Failed to write into SPI_0_CTRL\n");
		return ret;	
	}

	reg_value = supfpga_mm_reg16_read(PIO_IE_0);

	reg_value = (reg_value | spi_ie_n | SPI_GIE);
	/*Write to the interrupt control enable register0 to enable SPI interrupts from SPI controllers*/
	ret = supfpga_mm_reg16_write(reg_value, PIO_IE_0);// 16 bit write SPI_IE to address PIO_IE_0
	if (ret != 0) {
		printk("lmlsfpga_spi_setup(): Failed to write into interrupt control enable 0 register\n");
		return ret;
	}
	
	return 0;
}

static int lmlsfpga_spi_transfer(struct spi_device *spi, struct spi_message *msg)
{
	struct lmlsfpga_spi	*ls;
	unsigned long flags, spi_ctrl_offset, spi_tx_buf_offset, spi_rx_buf_offset, spi_len_offset;
	struct spi_transfer *xfer;
	int ret = 0;
	uint16_t reg_val = 0, reg_status = 0;

	spin_lock_irqsave(&ls->lock, flags);
	ls = spi_master_get_devdata(spi->master);

	if (unlikely(list_empty(&msg->transfers))){
		printk("lmlsfpga_spi_transfer(): No messaages.\n");
		spin_unlock_irqrestore(&ls->lock, flags);
		return -EINVAL;
	}

	/*Determine the SPI controller for SPI transfer*/
	switch(spi->master->bus_num) {
		case 3://SPI Controller 0
			spi_ctrl_offset = SPI_0_CTRL;
			spi_tx_buf_offset = SPI_TX0_BUF;
			spi_rx_buf_offset = SPI_RX0_BUF;
			spi_len_offset = SPI_0_LEN;
			break;
		case 4://SPI Controller 1
			spi_ctrl_offset = SPI_1_CTRL;
			spi_tx_buf_offset = SPI_TX1_BUF;
			spi_rx_buf_offset = SPI_RX1_BUF;
			spi_len_offset = SPI_1_LEN;
			break;
		case 5://SPI Controller 2
			spi_ctrl_offset = SPI_2_CTRL;
			spi_tx_buf_offset = SPI_TX2_BUF;
			spi_rx_buf_offset = SPI_RX2_BUF;
			spi_len_offset = SPI_2_LEN;
			break;
		case 6://SPI Controller 3
			spi_ctrl_offset = SPI_3_CTRL;
			spi_tx_buf_offset = SPI_TX3_BUF;
			spi_rx_buf_offset = SPI_RX3_BUF;
			spi_len_offset = SPI_3_LEN;
			break;
		default:
			printk("Unknown controller\n");
			spin_unlock_irqrestore(&ls->lock, flags);
			return -1;
	}

	list_for_each_entry(xfer, &msg->transfers, transfer_list) {
		const void *txbuf = xfer->tx_buf;
		void *rxbuf = xfer->rx_buf;
		int i;
#ifdef DEBUG_LMLSFPGA_SPI_DRV
		unsigned int byte_indx = 0;
#endif

#ifdef DBG_CAN_INTE_INTF_EFLG
		if ((*(const u8 *)(txbuf+1) == 0x2b) || 
				(*(const u8 *)(txbuf+1) == 0x2c) || 
				(*(const u8 *  )(txbuf+1) == 0x2d)) { //Debug for CANINTE, CANINTF and EFLG
			dev_dbg(&spi->dev, "0x%x: tx - \n", (*(const u8 *)(txbuf+1)));
		}
#endif

		for (i = 0; i < xfer->len; i= i+2) {
			uint16_t wrd = 0, wrd_tmp = 0;

			if (txbuf) {
				wrd = (uint16_t)(*(const u8 *)(txbuf+i));
				if (i+1 < xfer->len) {
					wrd_tmp = ((uint16_t)(*(const u8 *)(txbuf+i+1))) << 8;
				}
				wrd |= wrd_tmp;
			}

#ifdef DBG_CAN_INTE_INTF_EFLG
			if ((*(const u8 *)(txbuf+1) == 0x2b) || 
					(*(const u8 *)(txbuf+1) == 0x2c) || 
					(*(const u8 *)(txbuf+1) == 0x2d)) { //Debug for CANINTE, CANINTF and EFLG
				printk("0x%x\t", wrd);
			}
#endif

			/*Write to the SPI_TX_BUF register to transmit data on the SPI controller*/
			ret = supfpga_mm_reg16_write(wrd, spi_tx_buf_offset+i); //16 bit write
			if (ret != 0) {
				dev_dbg(&spi->dev, "lmlsfpga_spi_setup(): Failed to write into SFPGA TX buff register\n");
				spin_unlock_irqrestore(&ls->lock, flags);
				return ret;
			}
		}

		/*Read current SPI controller and SPI len register values*/
		reg_val = supfpga_mm_reg16_read(spi_len_offset);
		
		reg_val = reg_val & 0xFB00; // Take care not to clear MIS by writing 1 to it as MIS is Write 1 to clear and clear SPI_LEN register value.

        /*Assuming SPI_SIE is always 1 after setting it in setup*/
		reg_status = 0;
		reg_status = (uint16_t)(SPI_MIE | SPI_GO);
		if (reg_val & (reg_status << 8)) {
			dev_dbg(&spi->dev, "\nSPI controller, not ready for SPI transaction\n");
			spin_unlock_irqrestore(&ls->lock, flags);
			return -1;
		}
		
		reg_status = 0;
		reg_status = (uint16_t)SPI_GO;
		reg_val = reg_val | (reg_status << 8);
		reg_val = reg_val | ((uint16_t)xfer->len);
		ret = supfpga_mm_reg16_write(reg_val, spi_len_offset);
		if (ret != 0) {
			printk("lmlsfpga_spi_setup(): Failed to write into 0x%x to SPI control-length registers at address 0x%lx to start SPI transaction\n", reg_val, spi_len_offset);
			spin_unlock_irqrestore(&ls->lock, flags);
			return ret;
		}

#ifdef DBG_CAN_INTE_INTF_EFLG
		if ((*(const u8 *)(txbuf+1) == 0x2b) || 
				(*(const u8 *)(txbuf+1) == 0x2c) || 
				(*(const u8 *      )(txbuf+1) == 0x2d)) { //Debug for CANINTE, CANINTF and EFLG
			printk("\nSPI_CONFIG + SPI_LEN = 0x%x\n", reg_val);
		}
#endif

		/*Poll for the completion of SPI transfer*/
		reg_val = 0;
		/*Read current SPI controller register value*/
		reg_val = supfpga_mm_reg16_read(spi_len_offset);

		while((uint16_t)(reg_val & (reg_status << 8 ))) {
			udelay(5);
			reg_val = 0;
			/*Read current SPI controller register value*/
			reg_val = supfpga_mm_reg16_read(spi_len_offset);
		}

		reg_val = 0;

#ifdef DBG_CAN_INTE_INTF_EFLG
		if ((*(const u8 *)(txbuf+1) == 0x2b) || 
				(*(const u8 *)(txbuf+1) == 0x2c) || 
				(*(const u8 *      )(txbuf+1) == 0x2d)) { //Debug for CANINTE, CANINTF and EFLG
			dev_dbg(&spi->dev, "0x%x: rx - \n", (*(const u8 *)(txbuf+1)));
		}
#endif

		for (i = 0; i < xfer->len; i=i+2) {
			uint16_t wrd_tmp = 0;

			wrd_tmp = supfpga_mm_reg16_read(spi_rx_buf_offset+i);

#ifdef DBG_CAN_INTE_INTF_EFLG
			if ((*(const u8 *)(txbuf+1) == 0x2b) || 
					(*(const u8 *)(txbuf+1) == 0x2c) || 
					(*(const u8 *)(txbuf+1) == 0x2d)) { //Debug for CANINTE, CANINTF and EFLG
				printk("0x%x\t", wrd_tmp);
			}
#endif
			
			if (rxbuf) {
				*(u8 *)(rxbuf+i) = (u8)wrd_tmp;
				
				if (i+1 < xfer->len) {
					*(u8 *)(rxbuf+i+1) = (u8)(wrd_tmp >> 8);
				}
			}
		}

#ifdef DEBUG_LMLSFPGA_SPI_DRV
        if ((*(const u8 *)(txbuf+0) == 0x90) ||
                    (*(const u8 *)(txbuf+0) == 0x94)) {
			if (rx_pckt_defn) {
				rx_pckt_defn[rx_array_index].rx_bus_num = spi->master->bus_num;
				
				for (byte_indx = 0; byte_indx < xfer->len; byte_indx++) {
					if (rx_pckt_defn[rx_array_index].rx_frame) {
						rx_pckt_defn[rx_array_index].rx_frame[byte_indx] = *(u8 *)(rxbuf + byte_indx);
					}
				}
            }

			rx_array_index++;
			if (rx_array_index == MAX_DBG_ARRAY_SIZE) {
				rx_array_index = 0;
			}
			rx_pckt_cnt++;
        }
#endif

#ifdef DBG_CAN_INTE_INTF_EFLG
		if ((*(const u8 *)(txbuf+1) == 0x2b) || 
				(*(const u8 *)(txbuf+1) == 0x2c) || 
				(*(const u8 *  )(txbuf+1) == 0x2d)) { //Debug for CANINTE, CANINTF and EFLG
			printk("\n\n");
		}
#endif
	}

	msg->status = 0;
	msg->complete(msg->context);

	spin_unlock_irqrestore(&ls->lock, flags);

	return 0;
}

static void lmlsfpga_spi_cleanup(struct spi_device *spi)
{
	//struct lmlsfpga_spi	*ls = spi_master_get_devdata(spi->master);
#ifdef DEBUG_GENERAL_LMLSFPGA
	printk("In function lmlsfpga_spi_cleanup()\n");
#endif
}

/*-------------------------------------------------------------------------*/

static int __init lmlsfpga_spi_probe(struct platform_device *pdev)
{
	int			        ret;
	struct spi_master	*master;
	struct lmlsfpga_spi	*ls;

#ifdef DEBUG_GENERAL_LMLSFPGA
	printk ("In function lmlsfpga_spi_probe()\n");
#endif
	/* setup spi core then atmel-specific driver state */
	ret = -ENOMEM;
	master = spi_alloc_master(&pdev->dev, sizeof *ls);
	if (!master) {
		dev_info(&pdev->dev, "Unable to allocate memory for master \n");
		return -1;
	}

	master->bus_num = pdev->id;
	master->num_chipselect = 1;
	master->setup = lmlsfpga_spi_setup;
	master->transfer = lmlsfpga_spi_transfer;
	master->cleanup = lmlsfpga_spi_cleanup;
	platform_set_drvdata(pdev, master);

	ls = spi_master_get_devdata(master);

	ls->pdev = pdev;
	ret = spi_register_master(master);
	if (!ret)
		dev_info(&pdev->dev, "probed\n");

	return 0;
}

static int __exit lmlsfpga_spi_remove(struct platform_device *pdev)
{
	struct spi_master	*master = platform_get_drvdata(pdev);

#ifdef DEBUG_GENERAL_LMLSFPGA
	printk ("In function lmlsfpga_spi_remove()\n");
#endif
	spi_unregister_master(master);

	printk (KERN_ALERT "lmlsfpga_spi_remove(): spi_sfpga.ko unregistered");

#ifdef DEBUG_LMLSFPGA_SPI_DRV
	lmlsfpga_spi_sysfs_exit;
#endif

	return 0;
}

static struct platform_driver lmlsfpga_spi_driver = {
	.driver		= {
		.name	= "lmlsfpga_spi",
		.owner	= THIS_MODULE,
	},
	.remove		= __exit_p(lmlsfpga_spi_remove),
};

static int __init lmlsfpga_spi_init(void)
{
#ifdef DEBUG_GENERAL_LMLSFPGA
	printk ("In function lmlsfpga_spi_init()\n");
#endif

	/*Write 0x00 to clear the interrupt control enable register0*/
	if ((supfpga_mm_reg16_write(0x0000, PIO_IE_0)) != 0) {
		printk("lmlsfpga_spi_init(): Failed to write into interrupt control enable 0 register\n");
		return -1;
	}
	
	/*Write 0x00 to clear the SPI control and length register0*/ //To do 0x0400 to clear MIS?
	if ((supfpga_mm_reg16_write(0x0400, SPI_0_LEN)) != 0) {
		printk("lmlsfpga_spi_init(): Failed to write into interrupt control enable 0 register\n");
		return -1;
	}
	
	/*Write 0x00 to clear the SPI control and length register1*/  //To do 0x0400 to clear MIS?
	if ((supfpga_mm_reg16_write(0x0400, SPI_1_LEN)) != 0) {
		printk("lmlsfpga_spi_init(): Failed to write into interrupt control enable 0 register\n");
		return -1;
	}
	
	/*Write 0x00 to clear the SPI control and length register2*/  //To do 0x0400 to clear MIS?
	if ((supfpga_mm_reg16_write(0x0400, SPI_2_LEN)) != 0) {
		printk("lmlsfpga_spi_init(): Failed to write into interrupt control enable 0 register\n");
		return -1;
	}
	
	/*Write 0x00 to clear the SPI control and length register3*/  //To do 0x0400 to clear MIS?
	if ((supfpga_mm_reg16_write(0x0400, SPI_3_LEN)) != 0) {
		printk("lmlsfpga_spi_init(): Failed to write into interrupt control enable 0 register\n");
		return -1;
	}

#ifdef DEBUG_LMLSFPGA_SPI_DRV
	rx_pckt_cnt = 0;
	lmlsfpga_spi_sysfs_init();
#endif

	return platform_driver_probe(&lmlsfpga_spi_driver, lmlsfpga_spi_probe);
}
module_init(lmlsfpga_spi_init);

static void __exit lmlsfpga_spi_exit(void)
{
#ifdef DEBUG_GENERAL_LMLSFPGA
	printk ("In function lmlsfpga_spi_exit()\n");
#endif
	platform_driver_unregister(&lmlsfpga_spi_driver);

#ifdef DEBUG_LMLSFPGA_SPI_DRV
	lmlsfpga_spi_sysfs_exit();
#endif
}
module_exit(lmlsfpga_spi_exit);

MODULE_DESCRIPTION("LMLSFPGA SPI Controller driver");
MODULE_AUTHOR("Anudhan Sachidananda <anudhan.sachidananda@netapp.com>");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:LMLSFPGA_spi");
